#include "jannex/api/console.h"
#include "jannex/jannex.h"

#ifdef CONFIG_HTTP_SERVER_ENABLE_DEBUG_MODE

#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/semphr.h"
#include "freertos/event_groups.h"

#include "esp_websocket_client.h"
#include "esp_event.h"

#include <stdio.h>
#include <string.h>
#include <esp_log.h>
#include <esp_system.h>

#include "mbedtls/base64.h"

#define CALC_BASE64_ENCODED_LENGTH(input_length) ((4 * ((input_length + 2) / 3)) + 1)

char cacheMessage[CONFIG_HTTP_SERVER_DEBUG_CACHE_LENGTH];
char cacheMessageClone[CONFIG_HTTP_SERVER_DEBUG_CACHE_LENGTH];
int cacheMessageIndex = 0;

SemaphoreHandle_t cacheMutex;

const char *TAG = "Console";
const char *LOG_TAG = "[LOG]";

esp_websocket_client_handle_t client = NULL;

int skippedMessages = 0;

void sendOlderMessages()
{
    if (esp_websocket_client_is_connected(client))
    {
        // printf("cacheMessageIndex:%i\n", cacheMessageIndex);

        int newDataLength = 0;

        // mutex lock
        if (xSemaphoreTake(cacheMutex, (TickType_t)10) == pdTRUE)
        {
            if (cacheMessageIndex > 0)
            {
                // clone cacheMessage to cacheMessageClone
                memcpy(cacheMessageClone, cacheMessage, cacheMessageIndex);

                newDataLength = cacheMessageIndex;

                // clear cacheMessage
                cacheMessageIndex = 0;
            }

            // mutex unlock
            xSemaphoreGive(cacheMutex);
        }

        if (newDataLength > 0)
        {

            cacheMessageClone[newDataLength - 1] = '\0';
            esp_websocket_client_send_text(client, cacheMessageClone, newDataLength, 1000 / portTICK_PERIOD_MS);
            newDataLength = 0;

            /*// count \n in cacheMessage
            int count = 0;
            for (int i = 0; i < cacheMessageIndex; i++)
            {
                if (cacheMessage[i] == '\n')
                {
                    count++;
                }
            }

            printf("count:%i\n", count);

            // split cacheMessage into multiple messages
            char *pch = strtok(cacheMessage, "\n");
            while (pch != NULL)
            {
                if (strlen(pch) > 0)
                {
                    printf("pf-'%s'\n", pch);

                    esp_websocket_client_send_text(client, pch, strlen(pch), 1000 / portTICK_PERIOD_MS);
                }
                pch = strtok(NULL, "\n");
            }
            cacheMessageIndex = 0;*/
        }
        else
        {
        }
    }
    else
    {
    }
}

int jannex_LOG(const char *format, va_list args)
{
    int status = 0;

    //  sprintf to jannex_LOG func and dynamically allocate memory
    char *esp_message;
    status = vasprintf(&esp_message, format, args);

    size_t encoded_length = 0;
    size_t input_length = strlen(esp_message);
    size_t message_length = CALC_BASE64_ENCODED_LENGTH(input_length) + 1;

    char *message = (char *)malloc(message_length + 1 + 1); // + \n + \0

    if (message != NULL)
    {
        int ret = mbedtls_base64_encode((unsigned char *)message, message_length - 1, &encoded_length, (unsigned char *)esp_message, input_length);

        if (ret != 0)
        {
            printf("Base64 encoding failed!");
            skippedMessages++;

            goto END;
        }

        // set last char to \n
        message[encoded_length] = '\n';
        message[encoded_length + 1] = '\0';
    }
    else
    {
        printf("Memory allocation failed!");
        skippedMessages++;

        goto END;
    }

    message_length = encoded_length + 1;

    // check if socket is online
    /*if (client != NULL && esp_websocket_client_is_connected(client))
    {
        // print last char of message in hex
        // printf("last char: %02x\n", message[encoded_length - 1]);

        sendOlderMessages();

        printf("message_length:%i\n", message_length);
        esp_websocket_client_send_text(client, message, message_length, 1000 / portTICK_PERIOD_MS);
    }
    else*/
    {
        // Take the mutex before writing to the variable
        if (xSemaphoreTake(cacheMutex, portMAX_DELAY) == pdTRUE)
        {

            if (cacheMessageIndex + message_length + 1 > CONFIG_HTTP_SERVER_DEBUG_CACHE_LENGTH)
            {
                // Cache is full
                // replace at the end of the cache

                strcpy(cacheMessage + CONFIG_HTTP_SERVER_DEBUG_CACHE_LENGTH - 7, "!CCO!\n");
                cacheMessageIndex = CONFIG_HTTP_SERVER_DEBUG_CACHE_LENGTH;

                xSemaphoreGive(cacheMutex);
                goto END;
            }

            // cache message thread safe
            // xSemaphoreTake(cacheMessageSemaphore, portMAX_DELAY);

            // cache message
            if (cacheMessageIndex == 0)
            {
                strcpy(cacheMessage + cacheMessageIndex, message);
                cacheMessageIndex += message_length;
            }
            else
            {
                // cacheMessage[cacheMessageIndex] = '\n';
                strcpy(cacheMessage + cacheMessageIndex, message);
                cacheMessageIndex += message_length;
            }

            // Release the mutex
            xSemaphoreGive(cacheMutex);
        }
        else
        {
            // Unable to take the mutex
            // Handle the error
        }
    }

END:

    free(esp_message);
    free(message);

    if (status >= 0)
    {

        return vprintf(format, args); // print to console
    }
    return status;
}

static void log_error_if_nonzero(const char *message, int error_code)
{
    if (error_code != 0)
    {
        ESP_LOGE(TAG, "Last error %s: 0x%x", message, error_code);
    }
}

static void websocket_event_handler(void *handler_args, esp_event_base_t base, int32_t event_id, void *event_data)
{
    esp_websocket_event_data_t *data = (esp_websocket_event_data_t *)event_data;
    switch (event_id)
    {
    case WEBSOCKET_EVENT_CONNECTED:
        ESP_LOGI(TAG, "WEBSOCKET_EVENT_CONNECTED");
        sendOlderMessages();

        ESP_LOGI(TAG, "Websocket connected! %s", CONFIG_HTTP_SERVER_DEBUG_URL);
        break;
    case WEBSOCKET_EVENT_DISCONNECTED:
        ESP_LOGE(TAG, "Websocket disconnected!");
        ESP_LOGI(TAG, "WEBSOCKET_EVENT_DISCONNECTED");
        log_error_if_nonzero("HTTP status code", data->error_handle.esp_ws_handshake_status_code);
        if (data->error_handle.error_type == WEBSOCKET_ERROR_TYPE_TCP_TRANSPORT)
        {
            log_error_if_nonzero("reported from esp-tls", data->error_handle.esp_tls_last_esp_err);
            log_error_if_nonzero("reported from tls stack", data->error_handle.esp_tls_stack_err);
            log_error_if_nonzero("captured as transport's socket errno", data->error_handle.esp_transport_sock_errno);
        }
        break;
    case WEBSOCKET_EVENT_DATA:
        if (data->payload_len == 0)
        {
            break;
        }

        ESP_LOGI(TAG, "WEBSOCKET_EVENT_DATA");
        ESP_LOGI(TAG, "Received opcode=%d", data->op_code);
        if (data->op_code == 0x08 && data->data_len == 2)
        {
            ESP_LOGW(TAG, "Received closed message with code=%d", 256 * data->data_ptr[0] + data->data_ptr[1]);
        }
        else
        {
            ESP_LOGW(TAG, "Received=%.*s", data->data_len, (char *)data->data_ptr);
        }

        // If received data contains json structure it succeed to parse
        ESP_LOGW(TAG, "Total payload length=%d, data_len=%d, current payload offset=%d\r\n", data->payload_len, data->data_len, data->payload_offset);

        break;
    case WEBSOCKET_EVENT_ERROR:
        ESP_LOGI(TAG, "WEBSOCKET_EVENT_ERROR");
        log_error_if_nonzero("HTTP status code", data->error_handle.esp_ws_handshake_status_code);
        if (data->error_handle.error_type == WEBSOCKET_ERROR_TYPE_TCP_TRANSPORT)
        {
            log_error_if_nonzero("reported from esp-tls", data->error_handle.esp_tls_last_esp_err);
            log_error_if_nonzero("reported from tls stack", data->error_handle.esp_tls_stack_err);
            log_error_if_nonzero("captured as transport's socket errno", data->error_handle.esp_transport_sock_errno);
        }
        break;
    }
}

// function like printf and redirect the output to websocket

static void console_task(void *pvParameters)
{
    esp_websocket_client_config_t websocket_cfg = {
        .uri = CONFIG_HTTP_SERVER_DEBUG_URL,
        .port = CONFIG_HTTP_SERVER_DEBUG_PORT,
        .task_stack = 1024 * 6,
        .reconnect_timeout_ms = 5000,
        .network_timeout_ms = 500,
        .ping_interval_sec = 10000,
    };

    ESP_LOGI(TAG, "Connecting to %s... on port %d", websocket_cfg.uri, websocket_cfg.port);

    client = esp_websocket_client_init(&websocket_cfg);
    esp_websocket_register_events(client, WEBSOCKET_EVENT_ANY, websocket_event_handler, (void *)client);

    esp_websocket_client_start(client);

    // just test here some data
    /*char data[32];
    int i = 0;
    while (true)
    {
        int len = sprintf(data, "hello %04d", i++);
        ESP_LOGI(TAG, "Sending %s", data);

        vTaskDelay(1000 / portTICK_PERIOD_MS);
    }

    esp_websocket_client_destroy(client);*/

    while (true)
    {
        sendOlderMessages();
        vTaskDelay(500 / portTICK_PERIOD_MS);
    }
}

void initConsole()
{
    cacheMutex = xSemaphoreCreateMutex();
    if (cacheMutex == NULL)
    {
        // Mutex creation failed
        // Handle the error
        printf("initConsole Mutex creation failed!!!!!!!!!!!");
    }

    esp_log_set_vprintf(&jannex_LOG);

    if (cacheMutex == NULL)
    {
        // Mutex creation failed
        // Handle the error
        printf("initConsole Mutex creation failed!!!!!!!!!!!");
        ESP_LOGE(TAG, "initConsole Mutex creation failed!!!!!!!!!!!");
    }
}

void connectConsole()
{
    // wait unitl wifi is connected
    // xEventGroupWaitBits(s_wifi_event_group, NETWORK_CONNECTED_BIT, false, true, portMAX_DELAY);
    vTaskDelay(3000 / portTICK_PERIOD_MS);

    xTaskCreate(console_task, "console_task", 4096, NULL, 5, NULL);
}

#endif // CONFIG_HTTP_SERVER_ENABLE_DEBUG_MODE