#include <stdio.h>
#include "esp_log.h"
#include "esp_system.h"
#include "esp_event.h"
#include "esp_http_client.h"

#include "esp_crt_bundle.h"

static const char *TAG = "httpsRequest";

esp_err_t https_get_request_SYNC(const char *url, char *out_buffer, int *out_buffer_length, int out_buffer_size, int *status_code)
{
    int content_length = 0;

    esp_http_client_config_t config = {
        .url = url,
        .event_handler = NULL,
        .transport_type = HTTP_TRANSPORT_OVER_SSL,
        .disable_auto_redirect = false,
        .crt_bundle_attach = esp_crt_bundle_attach,

    };

    esp_http_client_handle_t client = esp_http_client_init(&config);
    if (client == NULL)
    {
        ESP_LOGE(TAG, "Failed to initialize HTTP client");
        return ESP_FAIL;
    }

    // GET Request
    esp_http_client_set_method(client, HTTP_METHOD_GET);
    esp_err_t err = esp_http_client_open(client, 0);
    if (err != ESP_OK)
    {
        ESP_LOGE(TAG, "Failed to open HTTP connection: %s", esp_err_to_name(err));
    }
    else
    {
        content_length = esp_http_client_fetch_headers(client);
        if (content_length < 0)
        {
            ESP_LOGE(TAG, "HTTP client fetch headers failed");
        }
        else
        {
            int data_read = esp_http_client_read_response(client, out_buffer, out_buffer_size);
            out_buffer[data_read] = '\0';
            *out_buffer_length = data_read;
            if (data_read >= 0)
            {
                *status_code = esp_http_client_get_status_code(client);
                /*ESP_LOGI(TAG, "HTTP GET Status = %d, content_length = %lld",
                         status_code,
                         esp_http_client_get_content_length(client));*/

                /*ESP_LOGI(TAG, "HTTP GET Response: %.*s", data_read, out_buffer);*/
            }
            else
            {
                ESP_LOGE(TAG, "Failed to read response");
            }
        }
    }

    esp_http_client_cleanup(client);
    return err;
}
