#include "jannex/jannex.h"

#include "jannex/otaAPI.h"
#include "jannex/otaUpdate.h"

#include "jannex/httpsRequest.h"
#include "json_parser.h"

#include "esp_system.h"

#include "esp_mac.h"

#include "esp_app_desc.h"
#include "esp_ota_ops.h"

#include "string.h"

// #include "esp_heap_trace.h"

// #define NUM_RECORDS 100
// static heap_trace_record_t trace_record[NUM_RECORDS]; // This buffer must be in internal RAM

bool updateAvailable = false;
bool isCheckingForUpdates = false;

void printFreeHeap(int start)
{

    if (start == 0)
    {
        ESP_LOGI(jannexTAG, " \\/ \\/ \\/ \\/ \\/ Free heap: ");
        ESP_LOGI(jannexTAG, " %i", heap_caps_get_free_size(MALLOC_CAP_8BIT));
    }
    else if (start == -1)
    {
        ESP_LOGI(jannexTAG, " %i", heap_caps_get_free_size(MALLOC_CAP_8BIT));
        ESP_LOGI(jannexTAG, " /\\ /\\ /\\ /\\ /\\ Free heap: ");
    }
    else
    {
        ESP_LOGI(jannexTAG, " %i", heap_caps_get_free_size(MALLOC_CAP_8BIT));
    }
}

void getNewestVersion(bool forceUpdate)
{

    isCheckingForUpdates = true;

    // buffer for the response
    size_t length = 256;

    char response[length];
    int status_code = 0;
    int responseLength = 0;

    // get mac address
    uint8_t mac[6];
    esp_efuse_mac_get_default(mac);

    // get current version
    const esp_app_desc_t *app_desc = esp_app_get_description();
    char *currentVersion = app_desc->version;

    char branch[32];
    getBranch(branch);

    char url[256];
    sprintf(url, "%s?mac=%02X:%02X:%02X:%02X:%02X:%02X&device=%s&branch=%s", OTA_URL_PATH_getNewestVersion, mac[0], mac[1], mac[2], mac[3], mac[4], mac[5], app_desc->project_name, branch);

    // make the request
    esp_err_t err = https_get_request_SYNC(url, response, &responseLength, length, &status_code);

    // ESP_LOGI(jannexTAG, "HTTP GET Response: %i %i %s", status_code, responseLength, response);

    jparse_ctx_t jctx;
    int ret = json_parse_start(&jctx, response, responseLength);
    if (ret != OS_SUCCESS)
    {
        ESP_LOGE(jannexTAG, "HTTP GET Response: %i %i %s", status_code, responseLength, response);
        ESP_LOGE(jannexTAG, "Parser failed");
        isCheckingForUpdates = false;

        goto FREE_MEM;
    }

    char newestVersion[32];

    if (json_obj_get_string(&jctx, "error", newestVersion, sizeof(newestVersion)) == OS_SUCCESS)
    {
        ESP_LOGW(jannexTAG, "Error: %s", newestVersion);
        // ERROR
        isCheckingForUpdates = false;

        goto FREE_MEM;
    }

    if (json_obj_get_string(&jctx, "version", newestVersion, sizeof(newestVersion)) == OS_SUCCESS)
    {
        // ESP_LOGI(jannexTAG, "newestVersion %s", newestVersion);
    }
    else
    {
        ESP_LOGI(jannexTAG, "No version");
        isCheckingForUpdates = false;

        goto FREE_MEM;
    }

    // print current version
    // ESP_LOGI(jannexTAG, "Current version: %s", currentVersion);

    // newestVersion and currentVersion are strings with the format two numbers separated by a dot (e.g. 1.2) now we need to compare them
    // first we need to split the strings into two numbers
    char newestVersionMajor[16];
    char newestVersionMinor[16];
    char currentVersionMajor[16];
    char currentVersionMinor[16];

    // split newestVersion
    int i = 0;
    while (newestVersion[i] != '.')
    {
        newestVersionMajor[i] = newestVersion[i];
        i++;
    }
    newestVersionMajor[i] = '\0';
    i++;
    int j = 0;
    while (newestVersion[i] != '\0')
    {
        newestVersionMinor[j] = newestVersion[i];
        i++;
        j++;
    }
    newestVersionMinor[j] = '\0';

    // split currentVersion
    i = 0;
    while (currentVersion[i] != '.')
    {
        currentVersionMajor[i] = currentVersion[i];
        i++;
    }
    currentVersionMajor[i] = '\0';
    i++;
    j = 0;
    while (currentVersion[i] != '\0')
    {
        currentVersionMinor[j] = currentVersion[i];
        i++;
        j++;
    }
    currentVersionMinor[j] = '\0';

    // compare the numbers
    int newestVersionMajorInt = atoi(newestVersionMajor);
    int newestVersionMinorInt = atoi(newestVersionMinor);
    int currentVersionMajorInt = atoi(currentVersionMajor);
    int currentVersionMinorInt = atoi(currentVersionMinor);

    if (newestVersionMajorInt != currentVersionMajorInt) // or only for new version: newestVersionMajorInt > currentVersionMajorInt
    {
        // major update
        ESP_LOGI(jannexTAG, "Major update available");
        updateAvailable = true;
    }
    else if (newestVersionMinorInt != currentVersionMinorInt) // or only for new version: newestVersionMinorInt > currentVersionMinorInt
    {
        ESP_LOGI(jannexTAG, "Minor update available");
        updateAvailable = true;
    }

    esp_partition_t *last_invalid_app = esp_ota_get_last_invalid_partition();
    if (last_invalid_app != NULL)
    {
        // get partition description

        esp_app_desc_t last_app_desc;
        esp_err_t ret = esp_ota_get_partition_description(last_invalid_app, &last_app_desc);
        if (ret == ESP_OK)
        {
            ESP_LOGI(jannexTAG, "Last invalid app partition: %s", last_app_desc.version);
            // check if the last invalid partition is the same as the newest version
            if (strcmp(last_app_desc.version, newestVersion) == 0)
            {
                // the last invalid partition is the same as the newest version -> DONT update
                ESP_LOGE(jannexTAG, "BAD UPDATE DETECTED -> KEEP CURRENT VERSION AND DO NOT UPDATE");
                vTaskDelay(10000 / portTICK_PERIOD_MS);
                updateAvailable = false;
            }
        }
        else
            ESP_LOGE(jannexTAG, "Could not get last invalid app partition description: %s", esp_err_to_name(ret));
    } /*
     else
     {
         ESP_LOGI(jannexTAG, "No invalid partition");
     }*/

    isCheckingForUpdates = false;

    // esp ota do update
    if (updateAvailable)
    {

        startOTAupdate(forceUpdate);
    }

FREE_MEM:

    json_parse_end(&jctx);
    // free(response); // crashes ESP
}

void printFreeHeapTask(void *parameter)
{
    while (1)
    {
        size_t heap = esp_get_free_heap_size();
        ESP_LOGI(jannexTAG, "Biggest free heap Block: %i", heap);
        ESP_LOGI(jannexTAG, "Total Free heap:         %i", heap_caps_get_free_size(MALLOC_CAP_8BIT));

        vTaskDelay(500 / portTICK_PERIOD_MS);
    }
}

void checkForUpdatesAndInstall(void *parameter)
{
    // ESP_ERROR_CHECK(heap_trace_init_standalone(trace_record, NUM_RECORDS));

    xTaskCreate(printFreeHeapTask, "printFreeHeapTask", 4096, NULL, 5, NULL);

    vTaskDelay(pdMS_TO_TICKS(5000)); // Delay for 5000 milliseconds
    while (1)
    {

        // get current ota state
        esp_partition_t *running = esp_ota_get_running_partition();
        esp_ota_img_states_t state;
        esp_err_t ret = esp_ota_get_state_partition(running, &state);

        if (ret == ESP_OK)
        {
            if (state == ESP_OTA_IMG_PENDING_VERIFY)
            {
                ESP_LOGI(jannexTAG, "OTA update pending verify...");
                // The firmware is half-way through an OTA update, but it's waiting to verify.
                // We'll just reboot.
                vTaskDelay(pdMS_TO_TICKS(5000)); // Delay for 5000 milliseconds
                continue;
            }
        }

        if (isCheckingForUpdates == false)
        {
            // ESP_ERROR_CHECK(heap_trace_start(HEAP_TRACE_LEAKS));
            getNewestVersion(false);
            // ESP_ERROR_CHECK(heap_trace_stop());
            // heap_trace_dump();
            // vTaskDelay(pdMS_TO_TICKS(50000)); // Delay for 50000 milliseconds
        }

#ifdef CONFIG_FIRMWARE_UPDATE_SCAN_SECONDS
        if (CONFIG_FIRMWARE_UPDATE_SCAN_SECONDS >= 1)
        {
            vTaskDelay(pdMS_TO_TICKS(CONFIG_FIRMWARE_UPDATE_SCAN_SECONDS * 1000)); // Delay for 1000 milliseconds
        }
        else
#endif
        {
            vTaskDelete(NULL);
        }
    }
}

void checkIfOTAwasSuccessful(void *parameter)
{
    esp_partition_t *running = esp_ota_get_running_partition();
    esp_ota_img_states_t state;
    esp_err_t ret = esp_ota_get_state_partition(running, &state);

    if (ret != ESP_OK)
    {
        ESP_LOGE(jannexTAG, "esp_ota_get_state_partition failed! %i", ret);
        vTaskDelete(NULL);
        return;
    }

    if (state == ESP_OTA_IMG_PENDING_VERIFY)
    {
        ESP_LOGI(jannexTAG, "ESP_OTA_IMG_PENDING_VERIFY");
        vTaskDelay(1000 * 30 / portTICK_PERIOD_MS); // wait 30sec minute
        // run diagnostic function ...
        bool diagnostic_is_ok = true; // diagnostic();
        if (diagnostic_is_ok)
        {
            ESP_LOGI(jannexTAG, "Diagnostics completed successfully! Continuing execution ...");
            esp_ota_mark_app_valid_cancel_rollback();
        }
        else
        {
            ESP_LOGE(jannexTAG, "Diagnostics failed! Start rollback to the previous version ...");
            esp_ota_mark_app_invalid_rollback_and_reboot();
        }
    }

    ESP_LOGI(jannexTAG, "State: %i", state);

    vTaskDelete(NULL);
}