#include "jannex/jannex.h"

#include "jannex/ota.h"
#include "jannex/otaAPI.h"

#include "esp_system.h"
#include "esp_mac.h"
#include "esp_log.h"

#include "esp_http_client.h"
#include "esp_crt_bundle.h"

#include "esp_ota_ops.h"
#include "esp_https_ota.h"

static const char *TAG = "otaUpgrade";

esp_err_t _http_event_handler(esp_http_client_event_t *evt)
{
    switch (evt->event_id)
    {
    case HTTP_EVENT_ERROR:
        ESP_LOGI(TAG, "HTTP_EVENT_ERROR");
        break;
    case HTTP_EVENT_ON_CONNECTED:
        ESP_LOGD(TAG, "HTTP_EVENT_ON_CONNECTED");
        break;
    case HTTP_EVENT_HEADER_SENT:
        ESP_LOGD(TAG, "HTTP_EVENT_HEADER_SENT");
        break;
    case HTTP_EVENT_ON_HEADER:
        ESP_LOGI(TAG, "HTTP_EVENT_ON_HEADER, key=%s, value=%s", evt->header_key, evt->header_value);
        break;
    case HTTP_EVENT_ON_DATA:
        ESP_LOGI(TAG, "HTTP_EVENT_ON_DATA, len=%d", evt->data_len);
        break;
    case HTTP_EVENT_ON_FINISH:
        ESP_LOGD(TAG, "HTTP_EVENT_ON_FINISH");
        break;
    case HTTP_EVENT_DISCONNECTED:
        ESP_LOGI(TAG, "HTTP_EVENT_DISCONNECTED");
        break;
    case HTTP_EVENT_REDIRECT:
        ESP_LOGI(TAG, "HTTP_EVENT_REDIRECT");
        break;
    }
    return ESP_OK;
}

void startOTAupdate(bool forceUpdate)
{
    // get mac address
    uint8_t mac[6];
    esp_efuse_mac_get_default(mac);

    // needed to get project name
    const esp_app_desc_t *app_desc = esp_app_get_description();

    char branch[32];
    getBranch(branch);

    char url[256];
    sprintf(url, "%s?mac=%02X:%02X:%02X:%02X:%02X:%02X&device=%s&branch=%s", OTA_URL_PATH_getBinary, mac[0], mac[1], mac[2], mac[3], mac[4], mac[5], app_desc->project_name, branch);

    esp_http_client_config_t config = {
        .url = url,
        .crt_bundle_attach = esp_crt_bundle_attach,
        .buffer_size = 4096 * 4,

        .event_handler = _http_event_handler,
        .keep_alive_enable = true,
#ifdef CONFIG_EXAMPLE_FIRMWARE_UPGRADE_BIND_IF // possible to bind to specific interface
        .if_name = &ifr,
#endif
    };

    esp_https_ota_config_t ota_config = {
        .http_config = &config,
    };

    ESP_LOGI(TAG, "Starting OTA update");
    ESP_LOGI(TAG, "Attempting to download update from %s", config.url);
    esp_err_t err = esp_https_ota(&ota_config);
    if (err == ESP_OK)
    {
        ESP_LOGI(TAG, "OTA Succeed, Rebooting...");
        esp_restart();
    }
    else
    {
        ESP_LOGE(TAG, "Firmware upgrade failed!");
    }
}
