#include "esp_system.h"
#include "esp_log.h"
#include "nvs_flash.h"

static const char *TAG = "Storage";

#ifndef GIT_DEFAULT_BRANCH
#define GIT_DEFAULT_BRANCH "unknown"
#endif

esp_err_t readStorage(const char *key, char *out_value, size_t *length, char *default_value)
{

    // read nvs for key
    nvs_handle_t my_handle;
    esp_err_t ret = nvs_open("storage", NVS_READONLY, &my_handle);
    if (ret != ESP_OK)
    {

        if (ret == ESP_ERR_NVS_NOT_FOUND)
        {
            ESP_LOGI(TAG, "The value is not initialized yet!\n");
            goto SET_DEFUALT_VALUE;
        }
        else
        {
            ESP_LOGE(TAG, "Error (%s) reading!\n", esp_err_to_name(ret));
        }

        return ret;
    }
    else
    {
        ret = nvs_get_str(my_handle, key, &out_value[0], &length);
        switch (ret)
        {
        case ESP_OK:
            // ESP_LOGI(TAG, "Current %s: %s", key, out_value);

            break;
        case ESP_ERR_NVS_NOT_FOUND:
            ESP_LOGI(TAG, "The value is not initialized yet!\n");

            break;
        default:
            ESP_LOGE(TAG, "Error (%s) reading!\n", esp_err_to_name(ret));
        }

        nvs_close(my_handle);
        if (ret == ESP_OK)
        {
            return ret;
        }
    }

SET_DEFUALT_VALUE:

    // if not found, set to default value
    ret = nvs_open("storage", NVS_READWRITE, &my_handle);
    if (ret != ESP_OK)
    {
        ESP_LOGE(TAG, "Error (%s) opening NVS handle!\n", esp_err_to_name(ret));
        return ret;
    }
    else
    {
        ret = nvs_set_str(my_handle, key, default_value);
        ESP_ERROR_CHECK(ret);
        ret = nvs_get_str(my_handle, key, &out_value[0], &length);
        ESP_ERROR_CHECK(ret);
        ESP_LOGI(TAG, "Current %s: %s", key, out_value);
        nvs_close(my_handle);
        return ret;
    }
}

esp_err_t getBranch(char *out_value)
{
    size_t length = 32;
    char default_value[32];
    esp_err_t ret = readStorage("branch", out_value, &length, GIT_DEFAULT_BRANCH);
    return ret;
}